"""
Module realize handlers mixin
"""

from typing import Awaitable, List, Optional, Union

from luna3.backport4.handler_objects import Policies
from luna3.common.http_objs import BinaryImage
from luna3.common.luna_response import LunaResponse
from luna3.common.requests import RequestPayload
from luna3.lunavl.mixins.base import MixinStub


class HandlersMixin(MixinStub):
    """Handlers Mixin to work with handlers."""

    def createHandler(
        self, policies: Policies = Policies(), description: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Create new handler. Handler determinate a list of rules for proccessing input images.
        You can set parameters for detecting faces, extracting attributes.
        You can set rules for matching attributes and creating faces from them.

        Args:
            policies: object with policies information
            description: user description

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with new handler id and url

        >>> from luna3.backport4.httpclient import LunaHttpClient
        >>> from luna3.backport4.handler_objects import DetectPolicy, ExtractPolicy, CreateFacePolicy, LinkToListPolicy
        >>> from luna3.backport4.handler_objects import MatchPolicy, Filters, MatchFilter, Policies
        >>>
        >>> client = LunaHttpClient(api=1)
        >>> detectPolicy = DetectPolicy(estimateHeadPose=1, estimateGaze=1, extractExif=1, rollThreshold=38)
        >>> extractPolicy = ExtractPolicy(extractBasicAttributes=1, extractDescriptor=1, scoreThreshold=0.7772)
        >>> matchPolicyFilters = [
        >>>     MatchPolicy(listId="cbf199a4-b27e-40f9-90d2-a4687c7c6b05",
        >>>                                    gender=1, ethnicities=[1, 2]),
        >>>     MatchPolicy(listId="25008fad-9e0f-451f-9c34-f9a453c539f2",
        >>>                                    ageGte=21, gender=0, ageLt=33),
        >>>     MatchPolicy(listId="cbf199a4-b27e-40f9-8776-a4687c7c6b05")
        >>> ]
        >>>
        >>> filters = Filters(gender=1, ethnicities=[3,4], ageLt=45,
        >>>                   matchesFilters=[
        >>>                       MatchFilter(listId="25008fad-9e0f-451f-9c34-f9a453c539f2",
        >>>                                   similarityGte=0.8),
        >>>                       MatchFilter(listId="cbf199a4-b27e-40f9-90d2-a4687c7c6b05"),
        >>>                   ])
        >>>
        >>> createFacePolicy = CreateFacePolicy(filters=filters, createFaces=1, setSampleAsAvatar=1)
        >>>
        >>> filters = Filters(ageGte=18, ageLt=25, matchesFilters=[
        >>>     MatchFilter("cbf199a4-b27e-40f9-8776-a4687c7c6b05", similarityGte=0.72, similarityLte=0.97)])
        >>> linkToListsPolicy = [
        >>>     LinkToListPolicy(listId="cbf199a4-b27e-40f9-8555-a4687c7c6b05"),
        >>>     LinkToListPolicy(listId="cbf199a4-b27e-40f9-8666-a4687c7c6b05", filters=filters),
        >>> ]
        >>>
        >>> policies = Policies(
        >>>     detectPolicy=detectPolicy, multifacePolicy=1, extractPolicy=extractPolicy,
        >>>     matchPolicy = matchPolicyFilters,
        >>>     createFacePolicy=createFacePolicy,
        >>>     linkToListsPolicy=linkToListsPolicy)
        >>> resp = client.createHandler(description="Strange persons", policies=policies)
        >>> print(resp.statusCode)
        201
        >>> print(resp.json)
        {"handler_id":"b72acaee-005d-43a6-a0b1-d9ba76fd0e1f",
         "url":"\\/1\\/handlers\\/b72acaee-005d-43a6-a0b1-d9ba76fd0e1f"}

        """
        data = {"policies": policies.asDict()}
        if description is not None:
            data["description"] = description

        payload = RequestPayload.buildJson(body=data)
        return self.makeRequest(f"{self.baseUri}/handlers", "POST", body=payload, **kwargs)

    def getHandlers(
        self, description: Optional[str] = None, page: int = 1, pageSize: int = 10, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get handlers by filters

        Args:
            description: Find all handlers with description which "like" of this parameter
            page: The number of page
            pageSize: The numbers of items on page

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse`: json with handlers info

        >>> from luna3.backport4.httpclient import LunaHttpClient
        >>>
        >>> client = LunaHttpClient()
        >>> resp = client.getHandlers(description="Strange Faces")
        >>> print(resp.statusCode)
        200
        >>> print(resp.json)
        [
          {
            "handler_id": "4201e801-79ac-447d-b900-a9021b31176f",
            "create_time": "2018-10-04T12:34:44.009245+03:00",
            "last_update_time": "2018-10-04T12:34:44.009245+03:00",
            "description": "Strange Faces 2017",
            "policies": {
              "detect_policy": {
                "estimate_head_pose": 1, "estimate_quality": 0, "estimate_gaze": 1,
                "estimate_eyes_attributes": 0, "detect_landmarks68": 0, "extract_exif": 1,
                "yaw_threshold": 180, "roll_threshold": 38, "pitch_threshold": 180
              },
              "multiface_policy": 1,
              "extract_policy": {
                "extract_basic_attributes": 1, "extract_descriptor": 1,  "score_threshold": 0.7772
              },
              "match_policy": [
                {
                  "list_id": "cbf199a4-b27e-40f9-90d2-a4687c7c6b22",
                  "filters": {"gender": 0, "age__lt": 33, "age__gte": 21}
                }
              ],
              "link_to_lists_policy": [
                {"list_id": "cbf199a4-b27e-40f9-8555-a4687c7c6b05"}
              ]
            }
          },
          {
            "handler_id": "c0702ff4-4674-4f1e-becc-1faa58ed6fa8",
            "create_time": "2018-10-04T12:34:15.809832+03:00",
            "last_update_time": "2018-10-04T12:34:15.809832+03:00",
            "description": "Strange Faces 2018",
            "policies": {
              "detect_policy": {
                "estimate_head_pose": 1, "estimate_quality": 0, "estimate_gaze": 1,
                "estimate_eyes_attributes": 0, "detect_landmarks68": 0, "extract_exif": 1,
                "yaw_threshold": 180, "roll_threshold": 38, "pitch_threshold": 180
              },
              "multiface_policy": 1,
              "extract_policy": {
                "extract_basic_attributes": 1,
                "extract_descriptor": 1,
                "score_threshold": 0.7772
              },
              "match_policy": [
                {
                  "list_id": "cbf199a4-b27e-40f9-90d2-a4687c7c6b22",
                  "filters": {"gender": 0, "age__lt": 33, "age__gte": 21}
                }
              ],
              "link_to_lists_policy": [
                {"list_id": "cbf199a4-b27e-40f9-8555-a4687c7c6b05"}
             ]
            }
          }
        ]
        >>> resp = client.getHandlers(description="Strange Faces 2019")
        >>> print(resp.statusCode)
        200
        >>> print(resp.json)
        []

        """
        queries = {"description": description, "page": page, "page_size": pageSize}
        return self.makeRequest(f"{self.baseUri}/handlers", "GET", queryParams=queries, **kwargs)

    def getHandlersCount(
        self, description: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get handlers count by filters

        Args:
            description: Find all handlers with description which "like" of this parameter

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse`:  handlers count

        >>> from luna3.backport4.httpclient import LunaHttpClient
        >>>
        >>> client = LunaHttpClient(accountId='02bbe8fe-7dd2-46bf-85c9-494693a9d30b')
        >>> resp = client.getHandlersCount(description="Strange Faces")
        >>> print(resp.statusCode)
        200
        >>> print(resp.json)
        5

        """
        queries = {"description": description}
        return self.makeRequest(f"{self.baseUri}/handlers/count", "GET", queryParams=queries, **kwargs)

    def getHandlerById(self, handlerId: str, **kwargs) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get handler by its id

        Args:
            handlerId: handler id (uuid4)

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse`: json with handler info

        >>> from luna3.backport4.httpclient import LunaHttpClient
        >>>
        >>> client = LunaHttpClient()
        >>> resp = client.getHandlerById(handlerId="4e127d4e-56ff-4ee7-9918-e45d7fd992dc")
        >>> print(resp.statusCode)
        200
        >>> print(resp.json)
        {
          "handler_id": "4e127d4e-56ff-4ee7-9918-e45d7fd992dc",
          "create_time": "2018-10-03T19:12:28.374842+03:00",
          "last_update_time": "2018-10-03T19:12:28.374842+03:00",
          "description": "test12",
          "policies": {
            "detect_policy": {
              "estimate_head_pose": 1,
              "estimate_quality": 0,
              "estimate_gaze": 1,
              "estimate_eyes_attributes": 0,
              "detect_landmarks68": 0,
              "extract_exif": 1,
              "yaw_threshold": 180,
              "roll_threshold": 38,
              "pitch_threshold": 180
            },
            "multiface_policy": 1,
            "extract_policy": {
              "extract_basic_attributes": 1,
              "extract_descriptor": 1,
              "score_threshold": 0.7772
            },
            "link_to_lists_policy": [
              {
                "list_id": "cbf199a4-b27e-40f9-8555-a4687c7c6b05"
              }
            ]
          }
        }

        """
        return self.makeRequest(f"{self.baseUri}/handlers/{handlerId}", "GET", **kwargs)

    def updateHandler(
        self, handlerId: str, policies: Policies = Policies(), description: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Update a handler. You can not update a part of handler.

        Args:
            handlerId: handler id
            policies: object with policies information
            description: user description

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return nothing in text field

        >>> from luna3.backport4.httpclient import LunaHttpClient
        >>> from luna3.backport4.handler_objects import DetectPolicy, ExtractPolicy, CreateFacePolicy, LinkToListPolicy
        >>> from luna3.backport4.handler_objects import MatchPolicy, Filters, MatchFilter, Policies
        >>>
        >>> client = LunaHttpClient(api=1)
        >>> detectPolicy = DetectPolicy(estimateHeadPose=1, estimateGaze=1, extractExif=1, rollThreshold=38)
        >>> extractPolicy = ExtractPolicy(extractBasicAttributes=1, extractDescriptor=1, scoreThreshold=0.7772)
        >>> matchPolicyList = [
        >>>     MatchPolicy(listId="cbf199a4-b27e-40f9-90d2-a4687c7c6b05", gender=1, ethnicities=[1, 2]),
        >>>     MatchPolicy(listId="25008fad-9e0f-451f-9c34-f9a453c539f2", ageGte=21, gender=0, ageLt=33),
        >>>     MatchPolicy(listId="cbf199a4-b27e-40f9-8776-a4687c7c6b05"),
        >>> ]
        >>>
        >>> filters = Filters(gender=1, ethnicities=[3,4], ageLt=45,
        >>>                     matchesFilters=[
        >>>                         MatchFilter(listId="cbf199a4-b27e-40f9-90d2-a4687c7c6b05"),
        >>>                         MatchFilter(listId="25008fad-9e0f-451f-9c34-f9a453c539f2",
        >>>                                     similarityGte=0.8)
        >>>                     ])
        >>> createFacePolicy = CreateFacePolicy(filters=filters, createFaces=1, setSampleAsAvatar=1)
        >>>
        >>> filters = Filters(ageGte=18, ageLt=25,
        >>>                   matchesFilters=[MatchFilter("cbf199a4-b27e-40f9-8776-a4687c7c6b05",
        >>>                                               similarityGte=0.72, similarityLte=0.97)])
        >>> linkToListsPolicy = [
        >>>     LinkToListPolicy(listId="cbf199a4-b27e-40f9-8555-a4687c7c6b05"),
        >>>     LinkToListPolicy(listId="cbf199a4-b27e-40f9-8666-a4687c7c6b05", filters=filters),
        >>> ]
        >>>
        >>> policies = Policies(
        >>>     detectPolicy=detectPolicy, multifacePolicy=1, extractPolicy=extractPolicy,
        >>>     matchPolicy = matchPolicyList,
        >>>     createFacePolicy=createFacePolicy,
        >>>     linkToListsPolicy=linkToListsPolicy)
        >>> resp = client.updateHandler(handlerId="cbf199a4-b27e-40f9-90d2-a4687c7c6b55",
        >>>                            description="Strange persons v1.2", policies=policies)
        >>> print(resp.statusCode)
        204
        >>> print(resp.text)

        """
        data = {"policies": policies.asDict()}
        if description is not None:
            data["description"] = description

        payload = RequestPayload.buildJson(body=data)
        return self.makeRequest(f"{self.baseUri}/handlers/{handlerId}", "PUT", body=payload, **kwargs)

    def checkHandlerById(self, handlerId: str, **kwargs) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Check whether handler exists by its id

        Args:
            handlerId: handler id (uuid4)

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse`: no text returns

        >>> from luna3.backport4.httpclient import LunaHttpClient
        >>>
        >>> client = LunaHttpClient()
        >>> resp = client.checkHandlerById(handlerId="4e127d4e-56ff-4ee7-9918-e45d7fd992dc")
        >>> print(resp.statusCode)
        200
        >>> print(resp.text)

        >>> resp = client.checkHandlerById(handlerId="4e127d4e-56ff-4ee7-9918-e45d7fd992dd")
        >>> print(resp.statusCode)
        404
        >>> print(resp.text)

        """
        return self.makeRequest(f"{self.baseUri}/handlers/{handlerId}", "HEAD", **kwargs)

    def deleteHandlerById(self, handlerId: str, **kwargs) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Remove handler by id

         Args:
             handlerId: handler id (uuid4)

         Returns:
             class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
             In body of :class: `~.LunaResponse`: no text returns

         >>> from luna3.backport4.httpclient import LunaHttpClient
         >>>
         >>> client = LunaHttpClient()
         >>> resp = client.deleteHandlerById(handlerId="f0a3e675-4b08-4856-83c6-934f79241be1")
         >>> print(resp.statusCode)
         204
         >>> print(resp.text)

         >>> resp = client.checkHandlerById(handlerId="4e127d4e-56ff-4ee7-9918-e45d7fd992dd")
         >>> print(resp.statusCode)
         404
         >>> print(resp.text)

        """
        return self.makeRequest(f"{self.baseUri}/handlers/{handlerId}", "DELETE", **kwargs)

    def emitEvents(
        self,
        handlerId: str,
        images: Union[BinaryImage, List[BinaryImage]] = None,
        urls: Optional[List[str]] = None,
        samples: Optional[List[str]] = None,
        externalId: Optional[str] = None,
        userData: Optional[str] = None,
        aggregateAttributes: Optional[int] = None,
        source: Optional[str] = None,
        tags: Optional[str] = None,
        warpedImage: int = None,
        lunaEventTime: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Emit input events.

        Args:
            handlerId: handler id
            images: Single image or images list (supported formats "jpeg, png, tif, ppm, bmp").
            urls: list of urls with images
            samples: list of images' ids
            externalId: external id
            userData: user data
            aggregateAttributes: aggregate all attributes (descriptot, age, gender) or not (0 or 1)
            source: source
            tags: tags
            warpedImage: Whether input image is a warped or arbitrary image.
            lunaEventTime: user defined event time

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json new handler id and url

        >>> from luna3.backport4.httpclient import LunaHttpClient
        >>>
        >>> client = LunaHttpClient(accountId='02bbe8fe-7dd2-46bf-85c9-494693a9d30b')
        >>> image = BinaryImage(STANDARD_IMAGE)
        >>> resp = client.emitEvents(handlerId, images=image)
        >>> print(resp.statusCode)
        201
        >>> print(resp.text)
        {
           "detections": {
               "failed_images":[],
               "succeeded_images":[
                   {"samples":
                       [
                           {
                               "sample_id":"75376113-b0ab-4a4e-99b0-c241ed90b73c",
                               "url":"\\/1\\/samples\\/75376113-b0ab-4a4e-99b0-c241ed90b73c",
                               "detection":{"rect":{"height":185,"width":133,"x":72,"y":41},
                               "landmarks5":[[23,71],[82,71],[42,113],[29,139],[79,138]]}}],
                               "filename":""
                           }
                       ]
                   },
               "events":[
                   {
                   "samples":["75376113-b0ab-4a4e-99b0-c241ed90b73c"],
                   "attributes":
                       {
                           "samples":["75376113-b0ab-4a4e-99b0-c241ed90b73c"],
                           "url":"\\/1\\/attributes\\/579a2712-8225-4e12-9195-c111f0ea00e4",
                           "attribute_id":"579a2712-8225-4e12-9195-c111f0ea00e4",
                           "basic_attributes":
                               {
                                   "age":27,
                                   "gender":0,
                                   "ethnicities":
                                       {
                                           "estimations":
                                               {
                                                   "african_american":0.0,
                                                   "asian":0.0,
                                                   "caucasian":1,
                                                   "indian":0.0
                                               },
                                           "predominant_ethnicity":"caucasian"
                                       }
                               },
                           "score":0.9413699508
                       },
                   "source":null,
                   "tags":null,
                   "event_id":"4b0d8150-1146-4f32-b043-9c57361c2ef0",
                   "url":"\\/1\\/events\\/4b0d8150-1146-4f32-b043-9c57361c2ef0",
                   "matches":null,
                   "face":null
               }
           ]
        }

        """
        if images is None and urls is None and samples is None:
            raise ValueError("Cannot find images, or samples, or urls in input args")

        queries = {
            "external_id": externalId,
            "user_data": userData,
            "aggregate_attributes": aggregateAttributes,
            "source": source,
            "tags": tags,
            "warped_image": warpedImage,
        }

        headers = {"Luna-Event-Time": lunaEventTime} if lunaEventTime is not None else {}
        if images:
            payload = RequestPayload(body=self.getBodyImages(headers, images), headers=headers)
        else:
            payload = RequestPayload.buildJson(
                body={key: value for key, value in (("urls", urls), ("samples", samples)) if value}, headers=headers
            )
        return self.makeRequest(
            f"{self.baseUri}/handlers/{handlerId}/events", "POST", queryParams=queries, body=payload, **kwargs
        )
