"""
Luna-faces client module
"""

from numbers import Real
from typing import Awaitable, Generator, List, Optional, Union

import aiohttp

from ..common.base_api import BaseHTTPAPI
from ..common.http_objs import RawDescriptor
from ..common.luna_response import LunaResponse
from ..common.requests import RequestPayload
from ..common.utils import getPreparedDescriptors


class FacesApi(BaseHTTPAPI):
    """
    Class for request to luna-faces.

    """

    def __init__(
        self,
        origin: Optional[str] = "http://127.0.0.1:5030",
        api: Optional[int] = 3,
        asyncRequest: Optional[bool] = False,
        lunaRequestId: Optional[str] = None,
        totalTimeout: Real = 60,
        connectTimeout: Optional[Real] = None,
        sockReadTimeout: Optional[Real] = None,
        sockConnectTimeout: Optional[Real] = None,
        *,
        session: aiohttp.ClientSession = None,
        echo: bool = False,
        **kwargs,
    ):
        super().__init__(
            origin=origin,
            api=api,
            asyncRequest=asyncRequest,
            lunaRequestId=lunaRequestId,
            totalTimeout=totalTimeout,
            connectTimeout=connectTimeout,
            sockReadTimeout=sockReadTimeout,
            sockConnectTimeout=sockConnectTimeout,
            session=session,
            echo=echo,
            **kwargs,
        )

    def createFace(
        self,
        accountId: str,
        attributeId: Optional[str] = None,
        basicAttributes: Optional[dict] = None,
        descriptors: Optional[List[Union[bytes, RawDescriptor]]] = None,
        descriptorSamples: List[str] = None,
        basicAttributesSamples: List[str] = None,
        eventId: Optional[str] = None,
        userData: Optional[str] = "",
        externalId: Optional[str] = None,
        avatar: Optional[str] = None,
        listIds: Optional[List[str]] = None,
        contentType: str = "application/msgpack",
        accept: str = "application/msgpack",
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Create new face in luna-faces

        Args:
            accountId: id of account
            attributeId: attribute id
            descriptors: list of descriptors different versions
            descriptorSamples: list of warp image id from which the attribute was extracted.
            basicAttributes: dict with basic attributes (age, gender, ethnicity)
            basicAttributesSamples: list of warp image id from which the basic attribute was extracted.
            eventId: reference to event which created face
            userData: face user data
            externalId: external id of the face, if it has its own mapping in the system
            avatar: face avatar url
            listIds: created face will be attach to these lists
            contentType: content type to use (supported now: "application/json", "application/msgpack" (default))
            accept: accept header ("application/json" or "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with created face id will be returned.

        """
        return self._saveFace(
            accountId=accountId,
            externalId=externalId,
            attributeId=attributeId,
            basicAttributes=basicAttributes,
            descriptors=descriptors,
            descriptorSamples=descriptorSamples,
            basicAttributesSamples=basicAttributesSamples,
            eventId=eventId,
            userData=userData,
            avatar=avatar,
            listIds=listIds,
            contentType=contentType,
            accept=accept,
            **kwargs,
        )

    def putFace(
        self,
        faceId: str,
        accountId: str,
        externalId: Optional[str] = None,
        attributeId: Optional[str] = None,
        basicAttributes: Optional[dict] = None,
        descriptors: Optional[List[Union[bytes, RawDescriptor]]] = None,
        descriptorSamples: List[str] = None,
        basicAttributesSamples: List[str] = None,
        eventId: Optional[str] = None,
        userData: Optional[str] = "",
        avatar: Optional[str] = None,
        listIds: Optional[List[str]] = None,
        contentType: str = "application/msgpack",
        accept: str = "application/msgpack",
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Put face in luna-faces

        Args:
            faceId: external face id
            accountId: id of account
            externalId: external id of the face, if it has its own mapping in the system
            attributeId: attribute id
            descriptors: list of descriptors different versions
            descriptorSamples: list of warp image id from which the attribute was extracted.
            basicAttributes: dict with basic attributes (age, gender, ethnicity)
            basicAttributesSamples: list of warp image id from which the basic attribute was extracted.
            eventId: reference to event which created face
            userData: face user data
            avatar: face avatar url
            listIds: created face will be attach to these lists
            contentType: content type to use (supported now: "application/json", "application/msgpack" (default))
            accept: accept header ("application/json" or "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with face id will be returned.

        """
        return self._saveFace(
            accountId=accountId,
            faceId=faceId,
            externalId=externalId,
            attributeId=attributeId,
            basicAttributes=basicAttributes,
            descriptors=descriptors,
            descriptorSamples=descriptorSamples,
            basicAttributesSamples=basicAttributesSamples,
            eventId=eventId,
            userData=userData,
            avatar=avatar,
            listIds=listIds,
            accept=accept,
            contentType=contentType,
            **kwargs,
        )

    def _saveFace(
        self,
        accountId: str,
        externalId: Union[str, None],
        attributeId: Union[str, None],
        basicAttributes: Union[dict, None],
        descriptors: Union[List[Union[bytes, RawDescriptor]], None],
        descriptorSamples: List[str],
        basicAttributesSamples: List[str],
        eventId: Union[str, None],
        userData: Union[str, None],
        avatar: Union[str, None],
        listIds: Union[List[str], None],
        contentType: str,
        accept: str,
        faceId: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Save face in luna-faces.

        Args:
            accountId: id of account
            faceId: external face id
            externalId: external id of the face, if it has its own mapping in the system
            attributeId: attribute id
            descriptors: list of descriptors different versions
            descriptorSamples: list of warp image id from which the attribute was extracted.
            basicAttributes: dict with basic attributes (age, gender, ethnicity)
            basicAttributesSamples: list of warp image id from which the basic attribute was extracted.
            eventId: reference to event which created face
            userData: face user data
            avatar: face avatar url
            listIds: created face will be attach to these lists
            contentType: content type to use (supported now: "application/json", "application/msgpack" (default))
            accept: accept header ("application/json" or "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with face id will be returned.

        """
        data = {}
        if attributeId is not None:
            data.setdefault("attribute", {})["attribute_id"] = attributeId
        if descriptors is not None:
            data.setdefault("attribute", {})["face_descriptors"] = getPreparedDescriptors(
                descriptors, encode=contentType != "application/msgpack"
            )
            if descriptorSamples is not None:
                data["attribute"]["face_descriptor_samples"] = descriptorSamples
        if basicAttributes is not None:
            data.setdefault("attribute", {})["basic_attributes"] = basicAttributes
            if basicAttributesSamples is not None:
                data["attribute"]["basic_attributes_samples"] = basicAttributesSamples
        if eventId is not None:
            data["event_id"] = eventId
        if userData is not None:
            data["user_data"] = userData
        if externalId is not None:
            data["external_id"] = externalId
        if avatar is not None:
            data["avatar"] = avatar
        if listIds is not None:
            data["lists"] = listIds

        if contentType == "application/msgpack":
            payload = RequestPayload.buildMsgpack(body=data)
        elif contentType == "application/json" or contentType is None:
            payload = RequestPayload.buildJson(body=data)
        else:
            raise ValueError(f"Unsupported Content-Type '{contentType}'")

        payload.applyAcceptHeader(accept)
        payload.applyAccountIdHeader(accountId)
        url = "{}/faces/{}".format(self.baseUri, faceId) if faceId is not None else "{}/faces".format(self.baseUri)
        method = "PUT" if faceId is not None else "POST"
        return self.makeRequest(url, method, body=payload, **kwargs)

    def getFaces(
        self,
        userData: Optional[str] = None,
        accountId: Optional[str] = None,
        eventId: Optional[str] = None,
        listId: Optional[str] = None,
        faceIds: Optional[List[str]] = None,
        timeLt: Optional[str] = None,
        timeGte: Optional[str] = None,
        page: Optional[int] = 1,
        pageSize: Optional[int] = 10,
        attributeIds: Optional[List[str]] = None,
        externalIds: Optional[List[str]] = None,
        faceIdGte: Optional[str] = None,
        faceIdLt: Optional[str] = None,
        targets: Optional[List[str]] = None,
        accept: str = "application/msgpack",
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get faces by filters. Optional you can set, that faces must belong to account.

        Args:
            userData: user data, part of user data, case insensitive
            accountId: account id
            eventId: event id
            listId: list id
            faceIds: face ids
            timeLt: upper bound of face create time
            timeGte: lower bound of face create time
            page: page count, default 1
            pageSize: page size, default 10, unlimited if -1
            attributeIds: attributes ids
            externalIds: list of external identifiers of persons, if they have their own mapping in the system
            faceIdGte: lower face id including boundary
            faceIdLt: upper face id excluding boundary
            targets: targets to get faces
            accept: accept header ("application/json" or "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with faces will be returned.

        """
        queries = {
            "user_data": userData,
            "account_id": accountId,
            "event_id": eventId,
            "list_id": listId,
            "face_ids": faceIds,
            "create_time__lt": timeLt,
            "create_time__gte": timeGte,
            "page": page,
            "page_size": pageSize,
            "external_ids": externalIds,
            "attribute_ids": attributeIds,
            "face_id__gte": faceIdGte,
            "face_id__lt": faceIdLt,
            "targets": targets,
        }
        headers = {"Accept": accept}
        return self.makeRequest(f"{self.baseUri}/faces", "GET", queryParams=queries, headers=headers, **kwargs)

    def getFacesCount(
        self,
        userData: Optional[str] = None,
        accountId: Optional[str] = None,
        eventId: Optional[str] = None,
        listId: Optional[str] = None,
        faceIds: Optional[List[str]] = None,
        timeLt: Optional[str] = None,
        timeGte: Optional[str] = None,
        attributeIds: Optional[List[str]] = None,
        externalIds: Optional[List[str]] = None,
        faceIdGte: Optional[str] = None,
        faceIdLt: Optional[str] = None,
        accept: str = "application/msgpack",
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Count faces by filters. Optional you can set, that faces must belong to account.

        Args:
            userData: user data, part of user data, case insensitive
            accountId: account id
            eventId: event id
            listId: list id
            faceIds: face ids
            timeLt: upper bound of face create time
            timeGte: lower bound of face create time
            attributeIds: attributes ids
            externalIds: list of external identifiers of persons, if they have their own mapping in the system
            faceIdGte: lower face id including boundary
            faceIdLt: upper face id excluding boundary
            accept: accept header ("application/json" or "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with number of faces will be returned.

        """
        queries = {
            "user_data": userData,
            "account_id": accountId,
            "event_id": eventId,
            "list_id": listId,
            "face_ids": faceIds,
            "create_time__lt": timeLt,
            "create_time__gte": timeGte,
            "external_ids": externalIds,
            "face_id__gte": faceIdGte,
            "face_id__lt": faceIdLt,
            "attribute_ids": attributeIds,
        }
        headers = {"Accept": accept}
        return self.makeRequest(f"{self.baseUri}/faces/count", "GET", queryParams=queries, headers=headers, **kwargs)

    def getFacesAttributesCount(
        self, accountId: Optional[str] = None, accept: str = "application/msgpack", **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Count faces with attributes. Optional you can set, that faces must belong to account.

        Args:
            accountId: account id
            accept: accept header ("application/json" or "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with number of faces will be returned.

        """
        queries = {"account_id": accountId}
        headers = {"Accept": accept}
        return self.makeRequest(
            f"{self.baseUri}/faces/attributes/count", "GET", queryParams=queries, headers=headers, **kwargs
        )

    def deleteFaces(
        self,
        faceIds: List[str],
        accountId: Optional[str] = None,
        ignore: Optional[int] = None,
        contentType: str = "application/msgpack",
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Delete faces. Optional you can set, that faces must belong to account.

        Args:
            faceIds: face id list
            accountId: account id
            ignore: ignore non-existing faces
            contentType: content type to use (supported now: "application/json", "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.

        """
        queries = {"account_id": accountId}
        if ignore is not None:
            queries["ignore"] = ignore

        data = {"face_ids": faceIds}
        if contentType == "application/msgpack":
            payload = RequestPayload.buildMsgpack(body=data)
        else:
            payload = RequestPayload.buildJson(body=data)
        return self.makeRequest(f"{self.baseUri}/faces", "DELETE", queryParams=queries, body=payload, **kwargs)

    def deleteFacesExtended(
        self,
        accountId: Optional[str] = None,
        targets: Optional[List[str]] = None,
        createTimeGte: Optional[str] = None,
        createTimeLt: Optional[str] = None,
        userData: Optional[str] = None,
        listId: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Delete faces. Optional you can set, that faces must belong to account.

        Args:
            accountId: faces account id
            targets: deletion info targets
            createTimeGte: lower including create time boundary
            createTimeLt: upper excluding create time boundary
            userData: user data
            listId: id of list where face in

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.

        """
        queries = {
            "account_id": accountId,
            "targets": targets,
            "create_time__gte": createTimeGte,
            "create_time__lt": createTimeLt,
            "user_data": userData,
            "list_id": listId,
        }

        return self.makeRequest(f"{self.baseUri}/faces/extended", "DELETE", queryParams=queries, **kwargs)

    def getFace(
        self,
        faceId: str,
        accountId: Optional[str] = None,
        targets: Optional[List[str]] = None,
        accept: str = "application/msgpack",
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get face by id. Optional you can set, that face must belong to account.

        Args:
            faceId: face id
            accountId: account id
            targets: targets to get face
            accept: accept header ("application/json" or "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with face will be returned.

        """
        queries = {"account_id": accountId, "targets": targets}
        headers = {"Accept": accept}
        return self.makeRequest(f"{self.baseUri}/faces/{faceId}", "GET", queryParams=queries, headers=headers, **kwargs)

    def checkFace(
        self, faceId: str, accountId: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Check face existence by id. Optional you can set, that face must belong to account.

        Args:
            faceId: face id
            accountId: account id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.

        """
        queries = {"account_id": accountId}
        return self.makeRequest(f"{self.baseUri}/faces/{faceId}", "HEAD", queryParams=queries, **kwargs)

    def updateFace(
        self,
        faceId: str,
        accountId: str,
        eventId: Optional[str] = None,
        userData: Optional[str] = None,
        externalId: Optional[str] = None,
        avatar: Optional[str] = None,
        contentType: str = "application/msgpack",
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Update face. You can update event id, user data, attribute id. Optional you can set, that face must belong
        to account.

        Args:
            faceId: face id
            accountId: account id
            eventId: new event id
            externalId: external id of the face, if it has its own mapping in the system
            userData: new user data
            avatar: face avatar url
            contentType: content type to use (supported now: "application/json", "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.

        """
        queries = {"account_id": accountId}

        data = {"event_id": eventId}
        if userData is not None:
            data["user_data"] = userData
        if externalId is not None:
            data["external_id"] = externalId
        if avatar is not None:
            data["avatar"] = avatar

        if contentType == "application/msgpack":
            payload = RequestPayload.buildMsgpack(body=data)
        else:
            payload = RequestPayload.buildJson(body=data)
        return self.makeRequest(f"{self.baseUri}/faces/{faceId}", "PATCH", queryParams=queries, body=payload, **kwargs)

    def deleteFace(
        self, faceId: str, accountId: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Delete face. Optional you can set, that face must belong to account.

        Args:
            faceId: face id
            accountId: account id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.

        """
        queries = {"account_id": accountId}
        return self.makeRequest(f"{self.baseUri}/faces/{faceId}", "DELETE", queryParams=queries, **kwargs)

    def createList(
        self,
        accountId: str,
        userData: Optional[str] = "",
        listId: Optional[str] = None,
        contentType: str = "application/msgpack",
        accept: str = "application/msgpack",
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Create list.

        Args:
            accountId: account id
            userData: user data
            listId: custom list id to create the list
            accept: accept header ("application/json" or "application/msgpack" (default))
            contentType: content type to use (supported now: "application/json", "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with created list id will be returned.

        """
        queries = {"list_id": listId}
        data = {"user_data": userData}
        if contentType == "application/msgpack":
            payload = RequestPayload.buildMsgpack(body=data)
        else:
            payload = RequestPayload.buildJson(body=data)
        payload.applyAcceptHeader(accept)
        payload.applyAccountIdHeader(accountId)
        return self.makeRequest(f"{self.baseUri}/lists", "POST", queryParams=queries, body=payload, **kwargs)

    def deleteLists(
        self,
        listIds: List[str],
        accountId: Optional[str] = None,
        withFaces: int = None,
        contentType: str = "application/msgpack",
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Delete lists. Optional you can set, that lists must belong to account.

        Args:
            listIds: ids of lists
            accountId: account id
            withFaces: remove lists and all the faces contained in the lists
            contentType: content type to use (supported now: "application/json", "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.

        """
        queries = {"account_id": accountId, "with_faces": withFaces}
        data = {"list_ids": listIds}
        if contentType == "application/msgpack":
            payload = RequestPayload.buildMsgpack(body=data)
        else:
            payload = RequestPayload.buildJson(body=data)
        return self.makeRequest(f"{self.baseUri}/lists", "DELETE", queryParams=queries, body=payload, **kwargs)

    def optionsLists(
        self,
        listIds: List[str],
        useParity: Optional[int] = None,
        contentType: str = "application/msgpack",
        accept: str = "application/msgpack",
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get lists with link and unlink keys by list ids.

        Args:
            listIds: lists ids
            useParity: if 1 - get even and odd max link/unlink keys else max link/unlink keys
            accept: accept header ("application/json" or "application/msgpack" (default))
            contentType: content type to use (supported now: "application/json", "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with lists will be returned.

        """
        queries = {"use_parity": useParity}
        data = {"list_ids": listIds}
        if contentType == "application/msgpack":
            payload = RequestPayload.buildMsgpack(body=data)
        else:
            payload = RequestPayload.buildJson(body=data)
        payload.applyAcceptHeader(accept)
        return self.makeRequest(f"{self.baseUri}/lists/linkkeys", "POST", queryParams=queries, body=payload, **kwargs)

    def getLists(
        self,
        userData: Optional[str] = None,
        userDataEq: Optional[str] = None,
        accountId: Optional[str] = None,
        createTimeLt: Optional[str] = None,
        createTimeGte: Optional[str] = None,
        lastUpdateTimeLt: Optional[str] = None,
        lastUpdateTimeGte: Optional[str] = None,
        listIds: Optional[List[str]] = None,
        listIdLt: Optional[str] = None,
        listIdGte: Optional[str] = None,
        page: Optional[int] = 1,
        pageSize: Optional[int] = 10,
        accept: str = "application/msgpack",
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get lists. Optional you can set, that lists must belong to account.

        Args:
            userData: user data, part of user data, case sensitive
            userDataEq: user data, case sensitive
            accountId: account id
            createTimeLt: upper bound of list create time
            createTimeGte: lower bound of list create time including boundary
            lastUpdateTimeLt: upper bound of list last update time
            lastUpdateTimeGte: lower bound of list last update time including boundary
            listIdLt: upper bound of list id
            listIdGte: lower bound of list id including boundary
            listIds: list ids
            page: page, default 1
            pageSize: page size, default 10
            accept: accept header ("application/json" or "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with lists will be returned.

        """
        queries = {
            "account_id": accountId,
            "user_data": userData,
            "user_data__eq": userDataEq,
            "page": page,
            "page_size": pageSize,
            "create_time__lt": createTimeLt,
            "create_time__gte": createTimeGte,
            "last_update_time__lt": lastUpdateTimeLt,
            "last_update_time__gte": lastUpdateTimeGte,
            "list_id__lt": listIdLt,
            "list_id__gte": listIdGte,
            "list_ids": listIds,
        }
        headers = {"Accept": accept}
        return self.makeRequest(f"{self.baseUri}/lists", "GET", queryParams=queries, headers=headers, **kwargs)

    def getListsCount(
        self,
        userData: Optional[str] = None,
        userDataEq: Optional[str] = None,
        accountId: Optional[str] = None,
        createTimeLt: Optional[str] = None,
        createTimeGte: Optional[str] = None,
        lastUpdateTimeLt: Optional[str] = None,
        lastUpdateTimeGte: Optional[str] = None,
        listIds: Optional[List[str]] = None,
        listIdLt: Optional[str] = None,
        listIdGte: Optional[str] = None,
        accept: str = "application/msgpack",
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get number of lists. Optional you can set, that lists must belong to account.

        Args:
            userData: user data, part of user data, case sensitive
            userDataEq: user data
            createTimeLt: upper bound of list create time
            createTimeGte: lower bound of list create time including boundary
            lastUpdateTimeLt: upper bound of list last update time
            lastUpdateTimeGte: lower bound of list last update time including boundary
            listIds: list ids
            listIdLt: upper bound of list id
            listIdGte: lower bound of list id including boundary
            accountId: account id
            accept: accept header ("application/json" or "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with number of lists will be returned.

        """
        queries = {
            "account_id": accountId,
            "user_data": userData,
            "list_ids": listIds,
            "list_id__lt": listIdLt,
            "list_id__gte": listIdGte,
            "create_time__lt": createTimeLt,
            "create_time__gte": createTimeGte,
            "last_update_time__lt": lastUpdateTimeLt,
            "last_update_time__gte": lastUpdateTimeGte,
            "user_data__eq": userDataEq,
        }
        headers = {"Accept": accept}
        return self.makeRequest(f"{self.baseUri}/lists/count", "GET", queryParams=queries, headers=headers, **kwargs)

    def getList(
        self,
        listId: str,
        accountId: Optional[str] = None,
        page: Optional[int] = 1,
        pageSize: Optional[int] = 10,
        accept: str = "application/msgpack",
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get list. Optional you can set, that list must belong to account.

        Args:
            listId: list id
            accountId: account id
            page: page count, default 1
            pageSize: page size, default 10
            accept: accept header ("application/json" or "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with list will be returned.

        """
        queries = {"page": page, "page_size": pageSize, "account_id": accountId}
        headers = {"Accept": accept}
        return self.makeRequest(f"{self.baseUri}/lists/{listId}", "GET", queryParams=queries, headers=headers, **kwargs)

    def checkList(
        self, listId: str, accountId: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Check list existence. Optional you can set, that list must belong to account.

        Args:
            listId: list id
            accountId: account id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.

        """
        queries = {"account_id": accountId}
        return self.makeRequest(f"{self.baseUri}/lists/{listId}", "HEAD", queryParams=queries, **kwargs)

    def deleteList(
        self, listId: str, accountId: Optional[str] = None, withFaces: int = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Remove list. Optional you can set, that list must belong to account.

        Args:
            listId: list id
            withFaces: remove lists and all faces witch is contained in the lists
            accountId: account id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.

        """
        queries = {"account_id": accountId, "with_faces": withFaces}
        return self.makeRequest(f"{self.baseUri}/lists/{listId}", "DELETE", queryParams=queries, **kwargs)

    def updateListUserData(
        self,
        listId: str,
        userData: str,
        accountId: Optional[str] = None,
        contentType: str = "application/msgpack",
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Update user data of list. Optional you can set, that list must belong to account.

        Args:
            listId: list id
            userData: new user data
            accountId: account id
            contentType: content type to use (supported now: "application/json", "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.

        """
        queries = {"account_id": accountId}
        data = {"user_data": userData}
        if contentType == "application/msgpack":
            payload = RequestPayload.buildMsgpack(body=data)
        else:
            payload = RequestPayload.buildJson(body=data)
        return self.makeRequest(f"{self.baseUri}/lists/{listId}", "PATCH", queryParams=queries, body=payload, **kwargs)

    def link(
        self,
        listId: str,
        faceIds: List[str],
        action: Optional[str] = "attach",
        accountId: Optional[str] = None,
        contentType: str = "application/msgpack",
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Attach or detach faces to list. Optional you can set, that list and faces must belong to account.

        Args:
            listId: list id
            faceIds: face ids
            action: attach or detach
            accountId: account id
            contentType: content type to use (supported now: "application/json", "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.

        """
        queries = {"account_id": accountId}
        data = {"action": action, "list_id": listId, "face_ids": faceIds}
        if contentType == "application/msgpack":
            payload = RequestPayload.buildMsgpack(body=data)
        else:
            payload = RequestPayload.buildJson(body=data)
        return self.makeRequest(f"{self.baseUri}/linker", "PATCH", queryParams=queries, body=payload, **kwargs)

    def getAttributesFromList(
        self,
        listId: str,
        accountId: Optional[str] = None,
        linkKeyLt: Optional[str] = None,
        linkKeyGte: Optional[str] = None,
        limit: Optional[int] = 1000,
        parity: Optional[int] = None,
        accept: str = "application/msgpack",
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get all attributes from list by period.

        Args:
            listId: list id
            accountId: account id to check list ownership
            linkKeyGte: upper including link key boundary
            linkKeyGte: lower including link key boundary
            limit: limit
            parity: if 1 - get with even and odd max link keys else max link key
            accept: accept header ("application/json" or "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with attributes will be returned.

        """
        queries = {
            "link_key__lt": linkKeyLt,
            "link_key__gte": linkKeyGte,
            "limit": limit,
            "account_id": accountId,
            "parity": parity,
        }
        headers = {"Accept": accept}
        return self.makeRequest(
            f"{self.baseUri}/lists/{listId}/attributes", "GET", queryParams=queries, headers=headers, **kwargs
        )

    def getDeletionsFromList(
        self,
        listId: str,
        accountId: Optional[str] = None,
        unlinkKeyLt: Optional[str] = None,
        unlinkKeyGte: Optional[str] = None,
        limit: Optional[int] = 1000,
        parity: Optional[int] = None,
        accept: str = "application/msgpack",
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get all deletions from list by period.

        Args:
            listId: list id
            accountId: account id to check list ownership
            unlinkKeyGte: upper including unlink key boundary
            unlinkKeyGte: lower including unlink key boundary
            limit: limit
            parity: if 1 - get with even and odd max unlink keys else max unlink key
            accept: accept header ("application/json" or "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with attributes will be returned.

        """
        queries = {
            "unlink_key__lt": unlinkKeyLt,
            "unlink_key__gte": unlinkKeyGte,
            "limit": limit,
            "account_id": accountId,
            "parity": parity,
        }
        headers = {"Accept": accept}
        return self.makeRequest(
            f"{self.baseUri}/lists/{listId}/deletions", "GET", queryParams=queries, headers=headers, **kwargs
        )

    def cleanLog(self, timeLt: Optional[str] = None, **kwargs) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Remove notes from log

        Args:
            timeLt: upper bound of face update time with timezone excluding boundary

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with attributes ids will be returned.

        """
        queries = {"time__lt": timeLt}
        return self.makeRequest(f"{self.baseUri}/linker/unlink_history", "PATCH", queryParams=queries, **kwargs)

    def facesAttributes(
        self, faceIds: List, accountId: Optional[str] = None, accept: str = "application/msgpack", **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Return attributeId of faces from faceIds list

        Args:
            faceIds: list of faceId for return their attributes
            accountId: account id
            accept: accept header ("application/json" or "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with attributes ids will be returned.

        """
        queries = {"faces_ids": faceIds, "account_id": accountId}
        headers = {"Accept": accept}
        return self.makeRequest(
            f"{self.baseUri}/faces/attributes", "GET", queryParams=queries, headers=headers, **kwargs
        )

    def getFaceAttributes(
        self,
        faceId: str,
        accountId: Optional[str] = None,
        targets: Optional[List[str]] = None,
        descriptorVersion: Optional[int] = None,
        accept: str = "application/msgpack",
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get face attributes.

        Args:
            faceId: face id
            accountId: account id
            targets: targets to get face attributes. Available targets: `create_time`, `face_descriptor`,
                     `basic_attributes`, `face_descriptor_samples`, `basic_attributes_samples`.
            descriptorVersion: descriptor version
            accept: accept header ("application/json" or "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with attributes will be returned.

        """
        queries = {"account_id": accountId, "descriptor_version": descriptorVersion, "targets": targets}
        headers = {"Accept": accept}
        return self.makeRequest(
            f"{self.baseUri}/faces/{faceId}/attributes", "GET", queryParams=queries, headers=headers, **kwargs
        )

    def _saveFaceAttributes(
        self,
        faceId: str,
        accountId: Union[str, None],
        attributeId: Union[str, None],
        basicAttributes: Union[dict, None],
        descriptors: Union[List[Union[bytes, RawDescriptor]], None],
        basicAttributesSamples: List[str],
        descriptorSamples: List[str],
        contentType: str,
        method: str,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Save face attributes.

        Args:
            faceId: face id
            accountId: account id
            attributeId: attribute id
            descriptors: list of descriptors different versions
            descriptorSamples: list of warp image id from which the attribute was extracted.
            basicAttributes: dict with basic attributes (age, gender, ethnicity)
            basicAttributesSamples: list of warp image id from which the basic attribute was extracted.
            contentType: content type to use (supported now: "application/json", "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse`

        """
        queries = {"account_id": accountId}

        data = {}
        if attributeId is not None:
            data["attribute_id"] = attributeId
        if descriptors is not None:
            data["face_descriptors"] = getPreparedDescriptors(descriptors, encode=contentType == "application/json")
            if descriptorSamples is not None:
                data["face_descriptor_samples"] = descriptorSamples
        if basicAttributes is not None:
            data["basic_attributes"] = basicAttributes
            if basicAttributesSamples is not None:
                data["basic_attributes_samples"] = basicAttributesSamples

        if contentType == "application/msgpack":
            payload = RequestPayload.buildMsgpack(body=data)
        else:
            payload = RequestPayload.buildJson(body=data)
        return self.makeRequest(
            f"{self.baseUri}/faces/{faceId}/attributes", method=method, queryParams=queries, body=payload, **kwargs
        )

    def putFaceAttributes(
        self,
        faceId: str,
        accountId: Optional[str] = None,
        attributeId: Optional[str] = None,
        basicAttributes: Optional[dict] = None,
        descriptors: Optional[List[Union[bytes, RawDescriptor]]] = None,
        basicAttributesSamples: List[str] = None,
        descriptorSamples: List[str] = None,
        contentType: str = "application/msgpack",
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Put face attributes.

        Args:
            faceId: face id
            accountId: account id
            attributeId: attribute id
            descriptors: list of descriptors different versions
            descriptorSamples: list of warp image id from which the attribute was extracted.
            basicAttributes: dict with basic attributes (age, gender, ethnicity)
            basicAttributesSamples: list of warp image id from which the basic attribute was extracted.
            contentType: content type to use (supported now: "application/json", "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse`

        """
        return self._saveFaceAttributes(
            faceId=faceId,
            accountId=accountId,
            attributeId=attributeId,
            basicAttributes=basicAttributes,
            basicAttributesSamples=basicAttributesSamples,
            descriptors=descriptors,
            descriptorSamples=descriptorSamples,
            contentType=contentType,
            method="PUT",
            **kwargs,
        )

    def updateFaceAttributes(
        self,
        faceId: str,
        accountId: Optional[str] = None,
        attributeId: Optional[str] = None,
        basicAttributes: Optional[dict] = None,
        descriptors: Optional[List[Union[bytes, RawDescriptor]]] = None,
        basicAttributesSamples: List[str] = None,
        descriptorSamples: List[str] = None,
        contentType: str = "application/msgpack",
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Update face attributes.

        Args:
            faceId: face id
            accountId: account id
            attributeId: attribute id
            descriptors: list of descriptors different versions
            descriptorSamples: list of warp image id from which the attribute was extracted.
            basicAttributes: dict with basic attributes (age, gender, ethnicity)
            basicAttributesSamples: list of warp image id from which the basic attribute was extracted.
            contentType: content type to use (supported now: "application/json", "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse`

        """
        return self._saveFaceAttributes(
            faceId=faceId,
            accountId=accountId,
            attributeId=attributeId,
            basicAttributes=basicAttributes,
            basicAttributesSamples=basicAttributesSamples,
            descriptors=descriptors,
            descriptorSamples=descriptorSamples,
            contentType=contentType,
            method="PATCH",
            **kwargs,
        )

    def deleteFaceAttributes(
        self, faceId: str, accountId: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Delete face attributes.

        Args:
            faceId: face id
            accountId: account id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse`

        """
        queries = {"account_id": accountId}
        return self.makeRequest(f"{self.baseUri}/faces/{faceId}/attributes", "DELETE", queryParams=queries, **kwargs)

    def getFaceAttributeSamples(
        self, faceId: str, accountId: Optional[str] = None, accept: str = "application/msgpack", **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get face attribute samples.

        Args:
            faceId: face id
            accountId: account id
            accept: accept header ("application/json" or "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with attribute samples will be returned.

        """
        queries = {"account_id": accountId}
        headers = {"Accept": accept}
        return self.makeRequest(
            f"{self.baseUri}/faces/{faceId}/attributes/samples", "GET", queryParams=queries, headers=headers, **kwargs
        )

    def getAttributeIds(
        self,
        listId: Optional[str] = None,
        accountId: Optional[str] = None,
        timeLt: Optional[str] = None,
        timeGte: Optional[str] = None,
        page: Optional[int] = 1,
        pageSize: Optional[int] = 1000,
        accept: str = "application/msgpack",
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get ids of attributes from list by period.

        Args:
            listId: list id
            accountId: account id
            timeLt: upper bound of face update time
            timeGte: lower bound of face update time
            page: page, default 1
            pageSize: page size, default 1000, max 100000
            accept: accept header ("application/json" or "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with attributes ids will be returned.

        """
        queries = {
            "time__lt": timeLt,
            "time__gte": timeGte,
            "page": page,
            "page_size": pageSize,
            "account_id": accountId,
            "list_id": listId,
        }
        headers = {"Accept": accept}
        return self.makeRequest(f"{self.baseUri}/attributes", "GET", queryParams=queries, headers=headers, **kwargs)

    def getAttributeIdsCount(
        self,
        listId: Optional[str] = None,
        accountId: Optional[str] = None,
        timeLt: Optional[str] = None,
        timeGte: Optional[str] = None,
        accept: str = "application/msgpack",
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get ids number of attributes from list by period.

        Args:
            listId: list id
            accountId: account id
            timeLt: upper bound of face update time
            timeGte: lower bound of face update time
            accept: accept header ("application/json" or "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with attributes ids will be returned.

        """
        queries = {"time__lt": timeLt, "time__gte": timeGte, "account_id": accountId, "list_id": listId}
        headers = {"Accept": accept}
        return self.makeRequest(
            f"{self.baseUri}/attributes/count", "GET", queryParams=queries, headers=headers, **kwargs
        )

    def getAttributes(
        self,
        attributeIds: List[str],
        descriptorVersion: int = None,
        targets: Optional[List[str]] = None,
        accountId: Optional[str] = None,
        accept: str = "application/msgpack",
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get list of attributes

        Args:
            attributeIds: list of attributes of interest
            targets: Available targets: `descriptor`, `basic_attributes`, `basic_attributes_samples`,
                `face_descriptors_samples`, `account_id`, `create_time`
            descriptorVersion: descriptor version to receive
            accountId: account id
            accept: accept header ("application/json" or "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with version.

        """
        queries = {
            "attribute_ids": attributeIds,
            "targets": targets,
            "account_id": accountId,
            "descriptor_version": descriptorVersion,
        }
        headers = {"Accept": accept}
        return self.makeRequest(f"{self.baseUri}/attributes", "GET", queryParams=queries, headers=headers, **kwargs)

    def getAttributesCount(
        self, accept: str = "application/msgpack", **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get count of temporary attributes

        Args:
            accept: accept header ("application/json" or "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with version.

        """
        headers = {"Accept": accept}
        return self.makeRequest(f"{self.baseUri}/attributes/count", "GET", headers=headers, **kwargs)

    def getAttributesReextract(
        self,
        accountId: Optional[str] = None,
        limit: Optional[int] = 100,
        attributesGte: Optional[str] = None,
        attributesLt: Optional[str] = None,
        accept: str = "application/msgpack",
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get list of attributes and their lists of sample ids

        Args:
            accountId: account id
            limit: limit of resulting list size
            attributesGte: lower including attribute id boundary
            attributesLt: upper excluding attribute id boundary
            accept: accept header ("application/json" or "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with version.

        """
        queries = {
            "account_id": accountId,
            "limit": limit,
            "attributes__gte": attributesGte,
            "attributes__lt": attributesLt,
        }
        headers = {"Accept": accept}
        return self.makeRequest(
            f"{self.baseUri}/attributes/reextract", "GET", queryParams=queries, headers=headers, **kwargs
        )

    def getAttributesCountReextract(
        self, accountId: Optional[str] = None, accept: str = "application/msgpack", **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get count of reextract descriptors' attrbiutes

        Args:
            accountId: account id
            accept: accept header ("application/json" or "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with version.

        """
        queries = {"account_id": accountId}
        headers = {"Accept": accept}
        return self.makeRequest(
            f"{self.baseUri}/attributes/reextract/count", "GET", queryParams=queries, headers=headers, **kwargs
        )

    def getDescriptorBatch(
        self,
        faceIds: Optional[List[str]] = None,
        listId: Optional[str] = None,
        linkKeyGte: Optional[int] = None,
        limit: Optional[int] = None,
        accountId: Optional[str] = None,
        checkExistence: int = 1,
        receiveExternalId: int = 0,
        parity: Optional[int] = None,
        contentType: str = "application/msgpack",
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get batch of descriptors (for matcher usage).
        Either receive descriptors by face ids (descriptors only) or by list id, linkKeyGte and limit
        (descriptors, link keys and uuids).

        Args:
            faceIds: face ids of descriptors to receive
            listId: list id to receive descriptors from
            linkKeyGte: lower including link key boundary
            limit: descriptors count to receive
            checkExistence: check faces and attributes existence or not
            accountId: account id
            receiveExternalId: receive external id or not
            parity: parity of link key
            contentType: content type to use (supported now: "application/json", "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with version.

        """
        queries = {
            "account_id": accountId,
            "check_existence": checkExistence,
            "parity": parity,
            "receive_external_id": receiveExternalId,
        }
        if faceIds is not None:
            data = {"face_ids": faceIds}
        else:
            data = {"list": {"list_id": listId, "link_key__gte": linkKeyGte, "limit": limit}}

        if contentType == "application/msgpack":
            payload = RequestPayload.buildMsgpack(body=data)
        else:
            payload = RequestPayload.buildJson(body=data)
        return self.makeRequest(
            f"{self.baseUri}/descriptors/batches", "POST", queryParams=queries, body=payload, **kwargs
        )

    def _saveAttribute(
        self,
        accountId: str,
        basicAttributes: Optional[dict],
        descriptors: Optional[List[Union[bytes, RawDescriptor]]],
        descriptorSamples: List[str],
        basicAttributesSamples: List[str],
        xpk: bytes,
        ttl: Optional[int],
        force: Optional[int],
        contentType: str,
        accept: str,
        attributeId: Optional[str] = None,
        **kwargs,
    ) -> Union[Generator[LunaResponse, None, None], LunaResponse]:
        """
        Create new attributes. Notice that XPK argument overwrites basic attributes and descriptors above.

        Args:
            accountId: account id of the attributes
            descriptors: list of descriptors different versions
            descriptorSamples: list of warp image id from which the attribute was extracted.
            basicAttributes: dict with basic attributes (age, gender, ethnicity)
            basicAttributesSamples: list of warp image id from which the basic attribute was extracted.
            xpk: SDKObject bytes to load attributes from
            ttl: temporary attribute ttl
            force: replace or not existence attribute
            contentType: content type to use (supported now: "application/json", "application/msgpack" (default))
            accept: accept header ("application/json" or "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with version.

        """
        meta = {}
        if descriptorSamples is not None:
            meta["face_descriptor_samples"] = descriptorSamples
        if basicAttributesSamples is not None:
            meta["basic_attributes_samples"] = basicAttributesSamples
        if xpk is not None:
            with aiohttp.MultipartWriter("form-data", self.multipartBoundary) as mpWriter:
                xpkPart = mpWriter.append(xpk, {"Content-Type": "application/x-vl-xpk"})
                xpkPart.set_content_disposition("attachment", name="xpk_file", filename="xpk-file", quote_fields=False)
                metaPart = mpWriter.append_json(meta, {"Content-Type": "application/json"})
                metaPart.set_content_disposition("attachment", name="meta", filename="meta", quote_fields=False)

                payload = RequestPayload.buildMultipart(body=mpWriter)
        else:
            body = meta
            if basicAttributes is not None:
                body["basic_attributes"] = basicAttributes
            if descriptors is not None:
                body["face_descriptors"] = getPreparedDescriptors(
                    descriptors, encode=contentType != "application/msgpack"
                )

            if contentType == "application/msgpack":
                payload = RequestPayload.buildMsgpack(body=body)
            else:
                payload = RequestPayload.buildJson(body=body)

        payload.applyAcceptHeader(accept)
        payload.applyAccountIdHeader(accountId)

        query = {"ttl": ttl, "force": force}
        url = f"{self.baseUri}/attributes/{attributeId}" if attributeId else f"{self.baseUri}/attributes"
        method = "PUT" if attributeId else "POST"
        return self.makeRequest(url, method, queryParams=query, body=payload, **kwargs)

    def createAttribute(
        self,
        accountId: str,
        basicAttributes: Optional[dict] = None,
        descriptors: Optional[List[Union[bytes, RawDescriptor]]] = None,
        descriptorSamples: List[str] = None,
        basicAttributesSamples: List[str] = None,
        xpk: bytes = None,
        ttl: Optional[int] = None,
        contentType: str = "application/msgpack",
        accept: str = "application/msgpack",
        **kwargs,
    ) -> Union[Generator[LunaResponse, None, None], LunaResponse]:
        """
        Create new attributes. Notice that XPK argument overwrites basic attributes and descriptors above.

        Args:
            accountId: account id of the attributes
            descriptors: list of descriptors different versions
            descriptorSamples: list of warp image id from which the attribute was extracted.
            basicAttributes: dict with basic attributes (age, gender, ethnicity)
            basicAttributesSamples: list of warp image id from which the basic attribute was extracted.
            xpk: SDKObject bytes to load attributes from
            ttl: temporary attribute ttl
            contentType: content type to use (supported now: "application/json", "application/msgpack" (default))
            accept: accept header ("application/json" or "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with attribute id.

        """
        return self._saveAttribute(
            accountId=accountId,
            ttl=ttl,
            basicAttributes=basicAttributes,
            basicAttributesSamples=basicAttributesSamples,
            descriptors=descriptors,
            descriptorSamples=descriptorSamples,
            xpk=xpk,
            force=None,
            contentType=contentType,
            accept=accept,
            **kwargs,
        )

    def putAttribute(
        self,
        attributeId: str,
        accountId: str,
        basicAttributes: Optional[dict] = None,
        descriptors: Optional[List[Union[bytes, RawDescriptor]]] = None,
        descriptorSamples: List[str] = None,
        basicAttributesSamples: List[str] = None,
        xpk: bytes = None,
        ttl: Optional[int] = None,
        force: Optional[int] = None,
        contentType: str = "application/msgpack",
        **kwargs,
    ) -> Union[Generator[LunaResponse, None, None], LunaResponse]:
        """
        Put attribute by id. Notice that XPK argument overwrites basic attributes and descriptors above.

        Args:
            attributeId: attribute id
            accountId: account id of the attributes
            descriptors: list of descriptors different versions
            descriptorSamples: list of warp image id from which the attribute was extracted.
            basicAttributes: dict with basic attributes (age, gender, ethnicity)
            basicAttributesSamples: list of warp image id from which the basic attribute was extracted.
            xpk: SDKObject bytes to load attributes from
            ttl: temporary attribute ttl
            force: replace or not existence attribute
            contentType: content type to use (supported now: "application/json", "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with version.

        """
        return self._saveAttribute(
            accountId=accountId,
            attributeId=attributeId,
            force=force,
            ttl=ttl,
            basicAttributes=basicAttributes,
            basicAttributesSamples=basicAttributesSamples,
            descriptors=descriptors,
            descriptorSamples=descriptorSamples,
            xpk=xpk,
            accept="application/json",
            contentType=contentType,
            **kwargs,
        )

    def checkAttribute(
        self, attributeId: str, accountId: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Check attributes with attribute id exists.

        Args:
            attributeId: attribute id
            accountId: account id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with version.

        """
        queries = {"account_id": accountId}
        return self.makeRequest(f"{self.baseUri}/attributes/{attributeId}", "HEAD", queryParams=queries, **kwargs)

    def getAttribute(
        self,
        attributeId: str,
        descriptorVersion: int = None,
        targets: Optional[List[str]] = None,
        accountId: Optional[str] = None,
        accept: str = "application/msgpack",
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Retrieve a attribute by id.

        Args:
            attributeId: attribute id
            targets: Available targets: `descriptor`, `basic_attributes`, `basic_attributes_samples`,
                `face_descriptors_samples`, `account_id`, `create_time`
            accountId: account id
            descriptorVersion: descriptor version to receive
            accept: accept header ("application/json" or "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with attribute.

        """
        queries = {"targets": targets, "account_id": accountId, "descriptor_version": descriptorVersion}
        headers = {"Accept": accept}
        return self.makeRequest(
            f"{self.baseUri}/attributes/{attributeId}", "GET", queryParams=queries, headers=headers, **kwargs
        )

    def getAttributeSamples(
        self, attributeId: str, accountId: Optional[str] = None, accept: str = "application/msgpack", **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Retrieve a attribute by id.

        Args:
            attributeId: attribute id
            accountId: account id
            accept: accept header ("application/json" or "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with samples.

        """
        queries = {"account_id": accountId}
        headers = {"Accept": accept}
        return self.makeRequest(
            f"{self.baseUri}/attributes/{attributeId}/samples", "GET", queryParams=queries, headers=headers, **kwargs
        )

    def deleteAttribute(
        self, attributeId: str, accountId: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Delete attribute with specifeid id

        Args:
            attributeId: attribute id
            accountId: account id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return  response empty body.

        """
        queries = {"account_id": accountId}
        return self.makeRequest(f"{self.baseUri}/attributes/{attributeId}", "DELETE", queryParams=queries, **kwargs)

    def getMissingDescriptors(
        self,
        missingVersion: int,
        accountId: Optional[str] = None,
        faceIds: Optional[List[str]] = None,
        faceIdGte: Optional[str] = None,
        faceIdLt: Optional[str] = None,
        limit: Optional[int] = None,
        contentType: str = "application/msgpack",
        accept: str = "application/msgpack",
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get face descriptors for re-extraction.

        Args:
            missingVersion: missing descriptors' version
            accountId: account id
            faceIds: list of face ids
            faceIdGte: lower face id including boundary
            faceIdLt: upper face id excluding boundary
            limit: maximum faces amount to return
            accept: accept header ("application/json" or "application/msgpack" (default))
            contentType: content type to use (supported now: "application/json", "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with data for re-extraction in the format:
                {"faces": [{"face_id": "<attribute_id>", "generation": <generation>, "samples": ["<sample_id>"]}]}

        """
        queries = {
            "account_id": accountId,
            "missing_version": missingVersion,
            "face_id__gte": faceIdGte,
            "face_id__lt": faceIdLt,
            "limit": limit,
        }
        payloadKwargs = {"body": {"face_ids": faceIds} if faceIds else None, "headers": {"Accept": accept}}
        if contentType == "application/msgpack":
            payload = RequestPayload.buildMsgpack(**payloadKwargs)
        else:
            payload = RequestPayload.buildJson(**payloadKwargs)
        return self.makeRequest(
            f"{self.baseUri}/faces/attributes/descriptors", "POST", queryParams=queries, body=payload, **kwargs
        )

    def getMissingDescriptorsCount(
        self,
        missingVersion: int,
        accountId: Optional[str] = None,
        faceIds: Optional[List[str]] = None,
        faceIdGte: Optional[str] = None,
        faceIdLt: Optional[str] = None,
        contentType: str = "application/msgpack",
        accept: str = "application/msgpack",
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get count of faces descriptors for re-extraction.

        Args:
            missingVersion: missing descriptors' version
            accountId: account id
            faceIds: list of face ids
            faceIdGte: lower face id including boundary
            faceIdLt: upper face id excluding boundary
            accept: accept header ("application/json" or "application/msgpack" (default))
            contentType: content type to use (supported now: "application/json", "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with descriptors for re-extraction count in the format:
                {"missing_descriptors_count": <missing_descriptors_count>}

        """
        queries = {
            "account_id": accountId,
            "missing_version": missingVersion,
            "face_id__gte": faceIdGte,
            "face_id__lt": faceIdLt,
        }
        payloadKwargs = {"body": {"face_ids": faceIds} if faceIds else None, "headers": {"Accept": accept}}
        if contentType == "application/msgpack":
            payload = RequestPayload.buildMsgpack(**payloadKwargs)
        else:
            payload = RequestPayload.buildJson(**payloadKwargs)
        return self.makeRequest(
            f"{self.baseUri}/faces/attributes/descriptors/count", "POST", queryParams=queries, body=payload, **kwargs
        )

    def getDescriptorCountInfo(
        self, accountId: Optional[str] = None, accept: str = "application/msgpack", **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get count of descriptors grouped by a descriptor version.

        Args:
            accountId: account id
            accept: accept header ("application/json" or "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with descriptor count in the format:
                {"descriptor_count_info": [{"descriptor_version": <descriptor_version>,
                                           "descriptor_count": <descriptor_count>}]}

        """
        queries = {"account_id": accountId}
        headers = {"Accept": accept}
        return self.makeRequest(
            f"{self.baseUri}/faces/attributes/descriptors/count", "GET", queryParams=queries, headers=headers, **kwargs
        )

    def getMissingBasicAttributes(
        self,
        accountId: Optional[str] = None,
        faceIds: Optional[List[str]] = None,
        faceIdGte: Optional[str] = None,
        faceIdLt: Optional[str] = None,
        limit: Optional[int] = None,
        contentType: str = "application/msgpack",
        accept: str = "application/msgpack",
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get face basic attributes for re-extraction.

        Args:
            accountId: account id
            faceIds: list of face ids
            faceIdGte: lower face id including boundary
            faceIdLt: upper face id excluding boundary
            limit: maximum faces amount to return
            accept: accept header ("application/json" or "application/msgpack" (default))
            contentType: content type to use (supported now: "application/json", "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with data for re-extraction in the format:
                {"faces": [{"face_id": "<attribute_id>", "generation": <generation>, "samples": ["<sample_id>"]}]}

        """
        queries = {"account_id": accountId, "face_id__gte": faceIdGte, "face_id__lt": faceIdLt, "limit": limit}
        payloadKwargs = {"body": {"face_ids": faceIds} if faceIds else None, "headers": {"Accept": accept}}
        if contentType == "application/msgpack":
            payload = RequestPayload.buildMsgpack(**payloadKwargs)
        else:
            payload = RequestPayload.buildJson(**payloadKwargs)
        return self.makeRequest(
            f"{self.baseUri}/faces/attributes/basic_attributes", "POST", queryParams=queries, body=payload, **kwargs
        )

    def getMissingBasicAttributesCount(
        self,
        accountId: Optional[str] = None,
        faceIds: Optional[List[str]] = None,
        faceIdGte: Optional[str] = None,
        faceIdLt: Optional[str] = None,
        contentType: str = "application/msgpack",
        accept: str = "application/msgpack",
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get count of face basic attributes for re-extraction.

        Args:
            accountId: account id
            faceIds: list of face ids
            faceIdGte: lower face id including boundary
            faceIdLt: upper face id excluding boundary
            accept: accept header ("application/json" or "application/msgpack" (default))
            contentType: content type to use (supported now: "application/json", "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with descriptors for re-extraction count in the format:
                {"missing_descriptors_count": <missing_descriptors_count>}

        """
        queries = {"account_id": accountId, "face_id__gte": faceIdGte, "face_id__lt": faceIdLt}
        payloadKwargs = {"body": {"face_ids": faceIds} if faceIds else None, "headers": {"Accept": accept}}
        if contentType == "application/msgpack":
            payload = RequestPayload.buildMsgpack(**payloadKwargs)
        else:
            payload = RequestPayload.buildJson(**payloadKwargs)
        return self.makeRequest(
            f"{self.baseUri}/faces/attributes/basic_attributes/count",
            "POST",
            queryParams=queries,
            body=payload,
            **kwargs,
        )

    def getBasicAttributesCountInfo(
        self, accountId: Optional[str] = None, accept: str = "application/msgpack", **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get count of basic attributes.

        Args:
            accountId: account id
            accept: accept header ("application/json" or "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with descriptor count in the format:
                {"basic_attributes_count": <basic_attributes_count>}

        """
        queries = {"account_id": accountId}
        headers = {"Accept": accept}
        return self.makeRequest(
            f"{self.baseUri}/faces/attributes/basic_attributes/count",
            "GET",
            queryParams=queries,
            headers=headers,
            **kwargs,
        )

    def deleteDescriptors(
        self, descriptorVersion: int, faceIdGte: Optional[str] = None, faceIdLt: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Delete descriptors by version.

        Args:
            descriptorVersion: descriptor version
            faceIdGte: lower face id including boundary
            faceIdLt: upper face id excluding boundary

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return list of attribute id for descriptors of preassigned version

        """
        queries = {"face_id__gte": faceIdGte, "face_id__lt": faceIdLt}
        return self.makeRequest(
            f"{self.baseUri}/faces/attributes/descriptors/{descriptorVersion}", "DELETE", queryParams=queries, **kwargs
        )

    def getListsDeletions(
        self,
        deletionTimeGte: Optional[str] = None,
        deletionTimeLt: Optional[str] = None,
        page: Optional[int] = None,
        pageSize: Optional[int] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get lists deletions log.

        Args:
            deletionTimeLt: upper bound of list deletion time
            deletionTimeGte: lower bound of list deletion time
            page: page count
            pageSize: page size

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return deletions lists log

        """
        queries = {
            "deletion_time__gte": deletionTimeGte,
            "deletion_time__lt": deletionTimeLt,
            "page": page,
            "page_size": pageSize,
        }
        return self.makeRequest(f"{self.baseUri}/lists/deletions", "GET", queryParams=queries, **kwargs)

    def clearListsDeletions(self, deletionTimeLt: str, **kwargs) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Clear lists deletions log.

        Args:
            deletionTimeLt: upper bound of list deletion time

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return empty response

        """
        queries = {"deletion_time__lt": deletionTimeLt}
        return self.makeRequest(f"{self.baseUri}/lists/deletions", "PATCH", queryParams=queries, **kwargs)
