"""
Module contains backward compatible common objects
"""

import base64
from dataclasses import dataclass
from typing import Any, Dict, Iterable, List, Optional, Union

from luna3.common.luna_request import guessContentType

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


BASE64_IMAGE_TYPES_MAP = {
    "image/jpeg": "image/x-jpeg-base64",
    "image/png": "image/x-png-base64",
    "image/bmp": "image/x-bmp-base64",
    "image/tiff": "image/x-tiff-base64",
    "image/x-portable-pixmap": "image/x-portable-pixmap-base64",
}


class BoundingBox:
    """
    Face bounding box

    Attributes:
        width (int): width
        height (int): height
        x (int): x coordinate of top left angle
        y (int): y coordinate of top left angle
    """

    __slots__ = ("x", "y", "width", "height")

    def __init__(self, width: int, height: int, x: int, y: int):  # pylint: disable-msg=invalid-name
        self.x: int = x  # pylint: disable=C0103
        self.y: int = y  # pylint: disable=C0103
        self.width: int = width
        self.height: int = height

    def asDict(self):
        """
        Get bounding box as dict
        Returns:
            dict with bounding box
        """
        return {"x": self.x, "y": self.y, "width": self.width, "height": self.height}


class IntThreshold:
    """
    Threshold with min and max values of int format

    Attributes:
        minimum: minimum
        maximum: maximum
    """

    def __init__(self, minimum: Optional[int] = None, maximum: Optional[int] = None):
        self.minimum = minimum
        self.maximum = maximum

    def asDict(self) -> Dict[str, int]:
        """Created dictionary representation"""
        result = {}
        if self.minimum is not None:
            result["min"] = self.minimum
        if self.maximum is not None:
            result["max"] = self.maximum
        return result


class FloatThreshold:
    """
    Threshold with min and max values of float format

    Attributes:
        minimum: minimum
        maximum: maximum
    """

    def __init__(self, minimum: Optional[float] = None, maximum: Optional[float] = None):
        self.minimum = minimum
        self.maximum = maximum

    def asDict(self) -> Dict[str, float]:
        """Created dictionary representation"""
        result = {}
        if self.minimum is not None:
            result["min"] = self.minimum
        if self.maximum is not None:
            result["max"] = self.maximum
        return result


class ImageFormatCheck:
    """
    Image format check structure
    Attributes:
        estimate: whether to enable check
        threshold: threshold
    """

    def __init__(self, estimate: bool = True, threshold: Optional[List[Literal["JPEG", "JPEG2000", "PNG"]]] = None):
        self.estimate = estimate
        self.threshold = threshold

    def asDict(self) -> Dict[str, Any]:
        """Created dictionary representation"""
        result = {"estimate": int(self.estimate)}
        if self.threshold is not None:
            result["threshold"] = self.threshold
        return result


class FloatQualityCheck:
    """
    Face quality check structure for check with float threshold
    Attributes:
        estimate: whether to enable check
        threshold: threshold
    """

    def __init__(self, estimate: bool = True, threshold: Optional[FloatThreshold] = None):
        self.estimate = estimate
        self.threshold = threshold

    def asDict(self) -> Dict[str, Any]:
        """Created dictionary representation"""
        result = {"estimate": int(self.estimate)}
        if self.threshold is not None:
            result["threshold"] = self.threshold.asDict()
        return result


class IntQualityCheck:
    """
    Face quality check structure for check with int threshold
    Attributes:
        estimate: whether to enable check
        threshold: threshold
    """

    def __init__(self, estimate: bool = True, threshold: Optional[IntThreshold] = None):
        self.estimate = estimate
        self.threshold = threshold

    def asDict(self) -> Dict[str, Any]:
        """Created dictionary representation"""
        result = {"estimate": int(self.estimate)}
        if self.threshold is not None:
            result["threshold"] = self.threshold.asDict()
        return result


class GlassesCheck:
    """
    Glasses check structure
    Attributes:
        estimate: whether to enable check
        threshold: threshold
    """

    def __init__(
        self, estimate: bool = True, threshold: Optional[List[Literal["no_glasses", "eyeglasses", "sunglasses"]]] = None
    ):
        self.estimate = estimate
        self.threshold = threshold

    def asDict(self) -> Dict[str, Any]:
        """Created dictionary representation"""
        result = {"estimate": int(self.estimate)}
        if self.threshold is not None:
            result["threshold"] = self.threshold
        return result


class LeftEyeCheck:
    """
    Left eye check structure
    Attributes:
        estimate: whether to enable check
        threshold: threshold
    """

    def __init__(self, estimate: bool = True, threshold: Optional[List[Literal["open", "occluded", "closed"]]] = None):
        self.estimate = estimate
        self.threshold = threshold

    def asDict(self) -> Dict[str, Any]:
        """Created dictionary representation"""
        result = {"estimate": int(self.estimate)}
        if self.threshold is not None:
            result["threshold"] = self.threshold
        return result


class RightEyeCheck:
    """
    Right eye check structure
    Attributes:
        estimate: whether to enable check
        threshold: threshold
    """

    def __init__(self, estimate: bool = True, threshold: Optional[List[Literal["open", "occluded", "closed"]]] = None):
        self.estimate = estimate
        self.threshold = threshold

    def asDict(self) -> Dict[str, Any]:
        """Created dictionary representation"""
        result = {"estimate": int(self.estimate)}
        if self.threshold is not None:
            result["threshold"] = list(self.threshold)
        return result


@dataclass
class Check:
    """
    Base check structure
    Attributes:
        estimate: whether to enable check
        threshold: threshold
    """

    estimate: bool = True
    threshold: Optional[Any] = None

    def asDict(self) -> Dict[str, Any]:
        """Created dictionary representation"""
        result = {"estimate": int(self.estimate)}
        if self.threshold is not None:
            result["threshold"] = self.threshold
        return result


@dataclass
class EyebrowsCheck(Check):
    """Check for eyebrows_type."""

    threshold: Iterable[Literal["squinting", "frowning", "raised", "neutral"]] = ("neutral",)


@dataclass
class ShouldersPositionCheck(Check):
    """Check for shoulders_position."""

    threshold: Iterable[Literal["non-parallel", "parallel", "hidden"]] = ("parallel",)


@dataclass
class SmileCheck(Check):
    """Check for smile_properties."""

    threshold: Iterable[Literal["none", "smile_lips", "smile_teeth"]] = ("none",)


@dataclass
class HeadwearCheck(Check):
    """Check for headwear_type."""

    threshold: Iterable[
        Literal[
            "none",
            "baseball_cap",
            "beanie",
            "peaked_cap",
            "shawl",
            "hat_with_earflaps",
            "helmet",
            "hood",
            "hat",
            "other",
        ]
    ] = ("none",)


@dataclass
class FaceColorCheck(Check):
    """Check for face_color_type."""

    threshold: Iterable[Literal["color", "grayscale", "infrared"]] = ("color",)


@dataclass
class FlagCheck(Check):
    """Universal check with 1/0 threshold."""

    threshold: Literal[0, 1] = 0


class FaceQualityChecks:
    """
    Face quality checks

    Attributes:
        imageFormat: image format check
        illuminationQuality: illumination quality check
        specularityQuality: specularity quality check
        blurrinessQuality: blurriness quality check
        darkQuality: dark quality check
        lightQuality: light quality check
        headYaw: head yaw check
        headPitch: head pitch check
        headRoll: head roll check
        gazeYaw: gaze yaw check
        gazePitch: gaze pitch check
        mouthSmiling: mouth smiling check
        mouthOccluded: mouth occluded check
        mouthOpen: mouth open check
        glasses: glasses check
        leftEye: left eye check
        rightEye: right eye check
        headHorizontalCenter: head horizontal center check
        headVerticalCenter: head vertical center check
        headWidth: head width check
        headHeight: head height check
        eyeDistance: eye distance check
        imageWidth: image width check
        imageHeight: image height check
        aspectRatio: aspect ratio check
        faceWidth: face width check
        faceHeight: face height check
        indentLeft: indent left check
        indentRight: indent right check
        indentUpper: indent upper check
        indentLower: indent lower check
        imageSize: image size check
        eyebrowsState: check configuration
        smileProperties: check configuration
        headwearType: check configuration
        naturalLight: check configuration
        radialDistortion: check configuration
        redEyes: check configuration
        faceColorType: check configuration
        illuminationUniformity: illumination uniformity check
        dynamicRange: dynamic range check
        backgroundLightness: face detection background lightness
        backgroundUniformity: face detection background uniformity
        shouldersPosition: shoulders position status
    """

    def __init__(
        self,
        imageFormat: Optional[ImageFormatCheck] = None,
        illuminationQuality: Optional[FloatQualityCheck] = None,
        specularityQuality: Optional[FloatQualityCheck] = None,
        blurrinessQuality: Optional[FloatQualityCheck] = None,
        darkQuality: Optional[FloatQualityCheck] = None,
        lightQuality: Optional[FloatQualityCheck] = None,
        headYaw: Optional[FloatQualityCheck] = None,
        headPitch: Optional[FloatQualityCheck] = None,
        headRoll: Optional[FloatQualityCheck] = None,
        gazeYaw: Optional[FloatQualityCheck] = None,
        gazePitch: Optional[FloatQualityCheck] = None,
        mouthSmiling: Optional[FloatQualityCheck] = None,
        mouthOccluded: Optional[FloatQualityCheck] = None,
        mouthOpen: Optional[FloatQualityCheck] = None,
        glasses: Optional[GlassesCheck] = None,
        leftEye: Optional[LeftEyeCheck] = None,
        rightEye: Optional[RightEyeCheck] = None,
        headHorizontalCenter: Optional[FloatQualityCheck] = None,
        headVerticalCenter: Optional[FloatQualityCheck] = None,
        headWidth: Optional[FloatQualityCheck] = None,
        headHeight: Optional[FloatQualityCheck] = None,
        eyeDistance: Optional[FloatQualityCheck] = None,
        imageWidth: Optional[IntQualityCheck] = None,
        imageHeight: Optional[IntQualityCheck] = None,
        aspectRatio: Optional[FloatQualityCheck] = None,
        faceWidth: Optional[IntQualityCheck] = None,
        faceHeight: Optional[IntQualityCheck] = None,
        indentLeft: Optional[IntQualityCheck] = None,
        indentRight: Optional[IntQualityCheck] = None,
        indentLower: Optional[IntQualityCheck] = None,
        indentUpper: Optional[IntQualityCheck] = None,
        imageSize: Optional[IntQualityCheck] = None,
        eyebrowsState: Optional[EyebrowsCheck] = None,
        smileProperties: Optional[SmileCheck] = None,
        headwearType: Optional[HeadwearCheck] = None,
        naturalLight: Optional[FlagCheck] = None,
        radialDistortion: Optional[FlagCheck] = None,
        redEyes: Optional[FlagCheck] = None,
        backgroundLightness: Optional[FloatQualityCheck] = None,
        backgroundUniformity: Optional[FloatQualityCheck] = None,
        faceColorType: Optional[FaceColorCheck] = None,
        illuminationUniformity: Optional[FloatQualityCheck] = None,
        dynamicRange: Optional[FloatQualityCheck] = None,
        shouldersPosition: Optional[ShouldersPositionCheck] = None,
    ):
        self.imageFormat = imageFormat
        self.illuminationQuality = illuminationQuality
        self.specularityQuality = specularityQuality
        self.blurrinessQuality = blurrinessQuality
        self.darkQuality = darkQuality
        self.lightQuality = lightQuality
        self.headYaw = headYaw
        self.headPitch = headPitch
        self.headRoll = headRoll
        self.gazeYaw = gazeYaw
        self.gazePitch = gazePitch
        self.mouthSmiling = mouthSmiling
        self.mouthOccluded = mouthOccluded
        self.mouthOpen = mouthOpen
        self.glasses = glasses
        self.leftEye = leftEye
        self.rightEye = rightEye
        self.headHorizontalCenter = headHorizontalCenter
        self.headVerticalCenter = headVerticalCenter
        self.headWidth = headWidth
        self.headHeight = headHeight
        self.eyeDistance = eyeDistance
        self.imageWidth = imageWidth
        self.imageHeight = imageHeight
        self.aspectRatio = aspectRatio
        self.faceWidth = faceWidth
        self.faceHeight = faceHeight
        self.indentLeft = indentLeft
        self.indentRight = indentRight
        self.indentUpper = indentUpper
        self.indentLower = indentLower
        self.imageSize = imageSize
        self.eyebrowsState = eyebrowsState
        self.smileProperties = smileProperties
        self.headwearType = headwearType
        self.naturalLight = naturalLight
        self.radialDistortion = radialDistortion
        self.redEyes = redEyes
        self.faceColorType = faceColorType
        self.illuminationUniformity = illuminationUniformity
        self.dynamicRange = dynamicRange
        self.backgroundLightness = backgroundLightness
        self.backgroundUniformity = backgroundUniformity
        self.shouldersPosition = shouldersPosition

    def asDict(self) -> Dict[str, Any]:
        """Creates dictionary representation"""
        result = {
            "image_format": self.imageFormat,
            "illumination_quality": self.illuminationQuality,
            "specularity_quality": self.specularityQuality,
            "blurriness_quality": self.blurrinessQuality,
            "dark_quality": self.darkQuality,
            "light_quality": self.lightQuality,
            "head_yaw": self.headYaw,
            "head_pitch": self.headPitch,
            "head_roll": self.headRoll,
            "gaze_yaw": self.gazeYaw,
            "gaze_pitch": self.gazePitch,
            "mouth_smiling": self.mouthSmiling,
            "mouth_occluded": self.mouthOccluded,
            "mouth_open": self.mouthOpen,
            "glasses": self.glasses,
            "left_eye": self.leftEye,
            "right_eye": self.rightEye,
            "head_horizontal_center": self.headHorizontalCenter,
            "head_vertical_center": self.headVerticalCenter,
            "head_width": self.headWidth,
            "head_height": self.headHeight,
            "eye_distance": self.eyeDistance,
            "image_width": self.imageWidth,
            "image_height": self.imageHeight,
            "aspect_ratio": self.aspectRatio,
            "face_width": self.faceWidth,
            "face_height": self.faceHeight,
            "indent_left": self.indentLeft,
            "indent_right": self.indentRight,
            "indent_upper": self.indentUpper,
            "indent_lower": self.indentLower,
            "image_size": self.imageSize,
            "eyebrows_state": self.eyebrowsState,
            "smile_properties": self.smileProperties,
            "headwear_type": self.headwearType,
            "natural_light": self.naturalLight,
            "radial_distortion": self.radialDistortion,
            "red_eyes": self.redEyes,
            "face_color_type": self.faceColorType,
            "illumination_uniformity": self.illuminationUniformity,
            "dynamic_range": self.dynamicRange,
            "background_lightness": self.backgroundLightness,
            "background_uniformity": self.backgroundUniformity,
            "shoulders_position": self.shouldersPosition,
        }
        return {key: value.asDict() for key, value in result.items() if value is not None}


class FaceQuality:
    """
    Class describes face quality structure

    Attributes:
        estimate: whether to estimate face quality
        enableFilter: whether to enable face quality filtering
        checks: face quality checks
    """

    def __init__(
        self, estimate: bool = False, enableFilter: bool = False, checks: FaceQualityChecks = FaceQualityChecks()
    ):
        self.estimate = estimate
        self.enableFilter = enableFilter
        self.checks = checks

    def asDict(self) -> Dict[str, Any]:
        """Creates dictionary representation"""
        result = {"estimate": int(self.estimate), "filter": int(self.enableFilter)}
        if self.checks is not None:
            result["checks"] = self.checks.asDict()
        return result


class BinaryData:
    """
    Binary Data class.

    Attributes:
        filename: filename.
        path: filename. may be path to the image.
        mimetype: property. image mimetype. Can be autodetected from body.
        _mimetype: private store mimetype
        body: property. binary representation of image. Can be uploaded by filename path.
        _body: private store body bytes
        multipartHeaders: Custom headers for multipart body

    """

    # Available mimetypes (should be defied in a derived class)
    availableMimetypes = tuple()

    def __init__(
        self,
        path: Optional[str] = None,
        filename: Optional[str] = None,
        mimetype: str = None,
        body: bytes = None,
        multipartHeaders: dict = None,
    ) -> None:
        """
        Init

        Args:
            path: path for load image from file
            filename: file name (is used for multipart ContentDisposition header)
            mimetype: mimetype
            body: binary body
            multipartHeaders: multipartHeaders

        """
        self.path = path
        self.filename = filename if filename is not None else path.split("/")[-1]
        self._body = body
        self._mimetype = mimetype

        if multipartHeaders is None:
            self.multipartHeaders = {}
        else:
            self.multipartHeaders = multipartHeaders

    @property
    def body(self) -> bytes:
        """
        Lazy body. If not set, uploads itself by the filename path.

        Returns:
             body
        """
        if self._body is not None:
            return self._body

        with open(self.path, "rb") as file:
            self._body = file.read()
        return self._body

    @body.setter
    def body(self, body: bytes):
        """
        Set body.

        Args:
            body: image bytes
        """
        self._body = body

    @property
    def mimetype(self) -> str:
        """
        Lazy mimetype. If not set, executed from body.

        Returns:
             mimetype
        """
        if self._mimetype is not None:
            return self._mimetype
        self._mimetype = self.getMimetype(self.body)
        return self._mimetype

    @mimetype.setter
    def mimetype(self, mimetype: str) -> None:
        """
        Set mimetype

        Args:
            image mimetype
        """
        self._mimetype = mimetype

    @property
    def multipartContentDispositionHeader(self) -> str:
        """
        Generate ContentDisposition header.

        Returns:
             Content-Disposition
        """
        return f'form-data; filename="{self.filename}"'

    @property
    def httpMultipartHeaders(self) -> Dict[str, str]:
        """
        Generate headers for each part of the multipart body.

        Returns:
             Content-Disposition and Content-Type headers.
        """
        return {
            "Content-Disposition": self.multipartContentDispositionHeader,
            "Content-Type": self.mimetype,
            **self.multipartHeaders,
        }

    def getMimetype(self, body: bytes) -> str:
        """ "
        The function gets mimetype of body.

        Args:
            body: binary data

        Raises:
            ValueError: if body has wrong mimetype

        Returns:
             Content-type of the body.
        """
        mimetype = guessContentType(body)
        if mimetype not in self.availableMimetypes:
            raise ValueError(f"Unsupported mimetype, availbale mimetypes are: {self.availableMimetypes}")
        return mimetype


class BinaryImageOrigin(BinaryData):
    """Binary Image Origin class."""

    availableMimetypes = ("image/jpeg", "image/png", "image/gif", "image/bmp", "image/tiff", "image/x-portable-pixmap")


class BinaryImage(BinaryData):
    """
    Binary Image class.

    Attributes:
        filename: filename.
        path: filename. may be path to the image.
        mimetype: property. image mimetype. Can be autodetected from body.
        _mimetype: private store mimetype
        body: property. binary representation of image. Can be uploaded by filename path.
        _body: private store body bytes
        multipartHeaders: Custom headers for multipart body
        detectTime (Optional[str]): image detection time in ISO format
        detectTs (Optional[float]): user-defined timestamp relative to something, such as the start of a video
        faceBoundingBoxList (Optional[List[BoundingBox]]): list of faces bounding boxes
        bodyBoundingBoxList (Optional[List[BoundingBox]]): list of human bodies bounding boxes
        imageOrigin (Optional[str]): image origin
        imageOriginMimetype (Optional[str]): image origin mimetype, used when imageOrigin is base64-encoded image
        meta (Optional[dict]): user-defined event metadata
    """

    availableMimetypes = ("image/jpeg", "image/png", "image/gif", "image/bmp", "image/tiff", "image/x-portable-pixmap")

    def __init__(
        self,
        path: Optional[str] = None,
        filename: Optional[str] = None,
        mimetype: str = None,
        body: bytes = None,
        multipartHeaders: dict = None,
        detectTime: Optional[str] = None,
        detectTs: Optional[float] = None,
        faceBoundingBoxList: Optional[List[BoundingBox]] = None,
        bodyBoundingBoxList: Optional[List[BoundingBox]] = None,
        trustedDetections: int = None,
        imageOrigin: Optional[Union[str, BinaryImageOrigin]] = None,
        imageOriginMimetype: Optional[str] = None,
        meta: Optional[dict] = None,
    ) -> None:
        """
        Init

        Args:
            path: path for load image from file
            filename: file name (is used for multipart ContentDisposition header)
            mimetype: mimetype
            body: binary body
            multipartHeaders: multipartHeaders
            detectTime: image detection time in ISO format
            detectTs: user-defined timestamp relative to something, such as the start of a video
            faceBoundingBoxList: list of faces bounding boxes
            bodyBoundingBoxList: list of bodies bounding boxes
            trustedDetections: whether detections were received with VL algorithms or not
            imageOrigin: image origin
            imageOriginMimetype: image origin mimetype, used when imageOrigin is base64-encoded image
            meta: user-defined event metadata
        """
        super().__init__(path=path, filename=filename, mimetype=mimetype, body=body, multipartHeaders=multipartHeaders)

        self.faceBoundingBoxList = faceBoundingBoxList
        self.bodyBoundingBoxList = bodyBoundingBoxList
        self.trustedDetections = trustedDetections
        self.detectTime = detectTime
        self.imageOrigin = imageOrigin
        self.imageOriginMimetype = imageOriginMimetype
        self.detectTs = detectTs
        self.meta = meta
        if self.trustedDetections is not None:
            self.multipartHeaders["X-Luna-Trusted-Detections"] = str(trustedDetections)

    def clone(self) -> "BinaryImage":
        """Clone instance"""
        return BinaryImage(
            filename=self.filename,
            path=self.path,
            mimetype=self._mimetype,
            body=self._body,
            multipartHeaders=self.multipartHeaders,
            detectTime=self.detectTime,
            detectTs=self.detectTs,
            faceBoundingBoxList=self.faceBoundingBoxList,
            bodyBoundingBoxList=self.bodyBoundingBoxList,
            trustedDetections=self.trustedDetections,
            imageOrigin=self.imageOrigin,
            imageOriginMimetype=self.imageOriginMimetype,
            meta=self.meta,
        )

    def asJson(self) -> dict:
        """
        Get Binary image as json with body decoded to str
        Returns:
            dict with body, mimetype and bounding box list if exists
        """
        if self.mimetype.endswith("base64"):
            image = self.body.decode()
            mimetype = self.mimetype
        else:
            image = base64.b64encode(self.body).decode()
            mimetype = BASE64_IMAGE_TYPES_MAP[self.mimetype]
        json = {"image": image, "mimetype": mimetype}
        if self.faceBoundingBoxList is not None:
            json["face_bounding_boxes"] = [boundingBox.asDict() for boundingBox in self.faceBoundingBoxList]
        if self.bodyBoundingBoxList is not None:
            json["body_bounding_boxes"] = [boundingBox.asDict() for boundingBox in self.bodyBoundingBoxList]
        if self.trustedDetections is not None:
            json["trusted_detections"] = self.trustedDetections
        if self.detectTime is not None:
            json["detect_time"] = self.detectTime
        if self.detectTs is not None:
            json["detect_ts"] = self.detectTs
        if self.imageOrigin is not None:
            json["image_origin"] = self.imageOrigin
        if self.imageOriginMimetype is not None:
            json["image_origin_mimetype"] = self.imageOriginMimetype
        if self.meta is not None:
            json["meta"] = self.meta
        return json


class BinaryDescriptorData(BinaryData):
    """Binary Descriptor Data class."""

    availableMimetypes = ("application/x-sdk-descriptor", "application/x-vl-xpk")


class PeopleCount:
    """
    Class describes people count estimation policy

    Attributes:
        estimate: whether to estimate people count
        peopleCoordinates: whether to return people coordinates
    """

    def __init__(self, estimate: bool = False, peopleCoordinates: bool = True):
        self.estimate = estimate
        self.peopleCoordinates = peopleCoordinates

    def asDict(self) -> Dict[str, Any]:
        """Creates dictionary representation"""
        result = {"estimate": int(self.estimate), "people_coordinates": int(self.peopleCoordinates)}
        return result
